mod packet;
mod utils;

use std::{
    fs::File,
    io::{self, Read},
    path::PathBuf,
    time::{Duration, Instant},
};

use clap::Parser;
use log::{debug, error, info, warn};
use serialport::SerialPort;
use skyward_mavlink::{
    lyra::MavMessage,
    mavlink::{write_v1_msg, MavHeader},
};

use crate::{
    packet::PacketSequence,
    utils::{get_first_stm32_serial_port, print_stats},
};

const ACK: u8 = 0x06;

#[derive(Debug, Parser)]
struct Cli {
    /// The serial port to use
    #[clap(short, long, value_name = "PORT")]
    port: Option<String>,

    /// The input csv file to read from
    csv_input: PathBuf,

    /// baud rate
    #[clap(short, long, default_value = "115200")]
    baud_rate: u32,
}

fn main() {
    simple_logger::SimpleLogger::new()
        .with_level(log::LevelFilter::Info)
        .with_local_timestamps()
        .env()
        .init()
        .unwrap();

    let args: Cli = Cli::parse();
    let mut packets = Vec::new();

    info!("reading from {}", args.csv_input.display());
    let mut rdr = csv::Reader::from_reader(File::open(args.csv_input).unwrap());
    for result in rdr.deserialize() {
        packets.push(MavMessage::PAYLOAD_FLIGHT_TM(result.unwrap()));
    }
    debug!("collected {} packets", packets.len());

    // convert to packet sequence data structure
    let mut packets = PacketSequence::from(packets);

    // Open the first serialport available.
    let port_name = if let Some(name) = args.port {
        name
    } else if let Some(name) = get_first_stm32_serial_port() {
        debug!("found STM32 device on {}", name);
        name
    } else {
        error!("FATAL: no STM32 serial port found, your device may not be connected properly");
        std::process::exit(1);
    };
    debug!("connecting to serial port {}...", port_name);
    let mut port = serialport::new(port_name, args.baud_rate)
        .open()
        .expect("Failed to open serial port");
    let mut write_port = port.try_clone().expect("Failed to clone");

    // let mut iter = packets.into_iter();

    // read from buffer and print to stdout
    let mut buffer = [0u8; 1000];
    let mut start: Option<Instant> = None;
    let mut times = Vec::new();

    // send packet if there is one
    while let Some(packet) = packets.wait_next() {
        if let Some(t) = start {
            times.push(t.elapsed().as_millis());
        }
        start = Some(Instant::now());
        write_port.write_packet(packet.to_owned());
        debug!(
            "sent packet ({}/{})",
            packets.sent_packets(),
            packets.total_packets()
        );

        // wait ACK
        loop {
            match port.read(&mut buffer) {
                Ok(t) => {
                    // eprint!("{}", String::from_utf8_lossy(&buffer[..t]));
                    // if didn't get the signal then print the buffer and exit with error
                    if buffer[..t].contains(&ACK) {
                        break;
                    } else {
                        warn!("didn't get the signal");
                        std::thread::sleep(Duration::from_millis(10));
                    }
                }
                Err(ref e) if e.kind() == io::ErrorKind::TimedOut => (),
                Err(e) => {
                    error!("FATAL: {:?}", e);
                    std::process::exit(1);
                }
            }
        }
        // eprintln!();
    }

    info!("sent {} packets", packets.sent_packets());
    print_stats(times);
}

trait MavLinkPort {
    fn write_packet(&mut self, packet: MavMessage);
}

impl<T: AsRef<dyn SerialPort> + std::io::Write> MavLinkPort for T {
    fn write_packet(&mut self, packet: MavMessage) {
        let header = MavHeader::default();
        write_v1_msg(self, header, &packet).expect("Failed to write to serial port");
    }
}
