use std::time::{Duration, Instant};

use serde::{Deserialize, Serialize};
use skyward_mavlink::{
    lyra::{MavMessage, ACK_TM_DATA, PAYLOAD_FLIGHT_TM_DATA},
    mavlink::Message,
};

pub trait TimedMessage: Message {
    fn timestamp(&self) -> Option<u64>;
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum PacketKind {
    Dummy(ACK_TM_DATA),
    PayloadFlightTm(PAYLOAD_FLIGHT_TM_DATA),
}

impl From<PacketKind> for MavMessage {
    fn from(value: PacketKind) -> Self {
        match value {
            PacketKind::Dummy(d) => MavMessage::ACK_TM(d),
            PacketKind::PayloadFlightTm(p) => MavMessage::PAYLOAD_FLIGHT_TM(p),
        }
    }
}

impl Default for PacketKind {
    fn default() -> Self {
        PacketKind::Dummy(ACK_TM_DATA {
            recv_msgid: 0,
            seq_ack: 0,
        })
    }
}

impl TimedMessage for MavMessage {
    fn timestamp(&self) -> Option<u64> {
        match self {
            MavMessage::PAYLOAD_FLIGHT_TM(data) => Some(data.timestamp),
            MavMessage::ACK_TM(_) => None,
            _ => None,
        }
    }
}

pub struct PacketSequence {
    packets: Vec<MavMessage>,
    index: usize,
    delta_from_start: u64,
    start_instant: Option<Instant>,
}

impl From<Vec<MavMessage>> for PacketSequence {
    fn from(value: Vec<MavMessage>) -> Self {
        let delta_from_start = value[0].timestamp().unwrap();
        Self {
            packets: value,
            index: 0,
            delta_from_start,
            start_instant: None,
        }
    }
}

impl PacketSequence {
    pub fn sent_packets(&self) -> usize {
        self.index
    }

    pub fn total_packets(&self) -> usize {
        self.packets.len()
    }

    pub fn wait_next(&mut self) -> Option<&MavMessage> {
        // return None if we reached the end of the sequence
        if self.index == self.packets.len() {
            return None;
        }

        // get time delta from the start
        let next_timestamp = self.packets[self.index].timestamp().unwrap();
        let time_delta = next_timestamp - self.delta_from_start;

        // get the instant the first packet was sent or set it if it's the first packet
        if let Some(start_inst) = self.start_instant {
            // get the instant to wait until
            let till = start_inst + Duration::from_micros(time_delta);

            // sleep for the difference between the last timestamp and the current one
            std::thread::sleep(till - Instant::now());
        } else {
            // set the start instant
            self.start_instant = Some(Instant::now());
        }
        self.index += 1;
        self.packets.get(self.index - 1)
    }
}
