use std::ops::{Deref, Range};

mod parser;

pub use parser::CachedLexer;

#[derive(Debug, Clone)]
pub struct Program(pub Vec<Blocks>);

#[derive(Debug, Clone)]
pub enum Blocks {
    TimeDefinition(TimeAnchorBlock),
    Generator(GeneratorBlock),
    Setting(SettingBlock),
}

#[derive(Debug, Clone)]
pub struct TimeAnchorBlock {
    span: Range<usize>,
    pub timeref: TimeRef,
    pub freq: Option<f64>,
}

#[derive(Debug, Clone)]
pub struct SettingBlock {
    span: Range<usize>,
    pub kind: SettingKind,
    pub settings: Vec<NumberParameter>,
}

impl SettingBlock {
    pub fn get(&self, name: &str) -> Option<&Number> {
        self.settings
            .iter()
            .find(|p| p.name == name)
            .map(|p| &p.value)
    }
}

#[derive(Debug, Clone)]
pub struct GeneratorBlock {
    span: Range<usize>,
    pub kind: GeneratorKind,
    pub parameters: Vec<ExprParameter>,
}

impl GeneratorBlock {
    pub fn get(&self, name: &str) -> Option<&TimeExpr> {
        self.parameters
            .iter()
            .find(|p| p.name == name)
            .map(|p| &p.value)
    }
}

#[derive(Debug, Clone)]
pub struct NumberParameter {
    span: Range<usize>,
    pub name: String,
    pub value: Number,
}

#[derive(Debug, Clone)]
pub struct ExprParameter {
    span: Range<usize>,
    name: String,
    pub value: TimeExpr,
}

#[derive(Debug, Clone)]
pub enum TimeExpr {
    Op(Box<Operation>),
    Num(Number),
    TimeVar(TimeVar),
    Group(GroupExpr),
}

impl TimeExpr {
    fn span(&self) -> Range<usize> {
        match self {
            TimeExpr::Op(op) => op.span.clone(),
            TimeExpr::Num(num) => num.span.clone(),
            TimeExpr::TimeVar(tv) => tv.span.clone(),
            TimeExpr::Group(expr) => expr.span.clone(),
        }
    }
}

#[derive(Debug, Clone)]
pub struct GroupExpr {
    span: Range<usize>,
    pub value: Box<TimeExpr>,
}

#[derive(Debug, Clone)]
pub struct Operation {
    span: Range<usize>,
    pub op: Operator,
    pub lhs: Box<TimeExpr>,
    pub rhs: Box<TimeExpr>,
}

#[derive(Debug, Clone)]
pub enum Operator {
    Add,
    Sub,
    Mul,
    Div,
    Pow,
}

#[derive(Debug, Clone)]
pub struct Number {
    span: Range<usize>,
    pub value: f64,
}

impl Deref for Number {
    type Target = f64;

    fn deref(&self) -> &Self::Target {
        &self.value
    }
}

#[derive(Debug, Clone)]
pub struct Generator {
    span: Range<usize>,
    kind: GeneratorKind,
}

#[derive(Debug, Clone)]
pub enum GeneratorKind {
    PositionParametric,
}

#[derive(Debug, Clone)]
pub struct Setting {
    span: Range<usize>,
    name: SettingKind,
}

#[derive(Debug, Clone)]
pub enum SettingKind {
    GPS,
}

#[derive(Debug, Clone)]
pub struct TimeRef {
    span: Range<usize>,
    offset: f64,
}

#[derive(Debug, Clone)]
pub struct TimeStepFreq {
    span: Range<usize>,
    freq: f64,
}

impl Deref for TimeRef {
    type Target = f64;

    fn deref(&self) -> &Self::Target {
        &self.offset
    }
}

#[derive(Debug, Clone)]
pub struct TimeVar {
    span: Range<usize>,
}
