use std::{collections::VecDeque, ops::Range};

use logos::{Logos, SpannedIter};

use super::*;
use crate::tokens::Token;

trait Lookahead: Iterator {
    type LookedItem;

    fn lookahead(&mut self, times: usize) -> Option<Self::LookedItem>;
    fn digest(&mut self, times: usize);
}

#[derive(Clone)]
pub struct CachedLexer<'a> {
    iter: SpannedIter<'a, Token>,
    cached: VecDeque<(Token, Range<usize>)>,
}

impl Iterator for CachedLexer<'_> {
    type Item = (Token, Range<usize>);

    fn next(&mut self) -> Option<Self::Item> {
        self.digest(1);
        self.lookahead(0)
    }
}

impl<'a> Lookahead for CachedLexer<'a> {
    type LookedItem = Self::Item;

    fn digest(&mut self, times: usize) {
        let to_pop = times.min(self.cached.len());
        for _ in 0..to_pop {
            self.cached.pop_front();
        }
        for _ in to_pop..times {
            self.iter.next();
        }
    }

    fn lookahead(&mut self, times: usize) -> Option<Self::LookedItem> {
        while self.cached.len() <= times {
            let (token, range) = self.iter.next()?;
            self.cached.push_back((token.unwrap(), range));
        }
        Some(self.cached[times].clone())
    }
}

impl<'a> CachedLexer<'a> {
    pub fn new(input: &'a str) -> Self {
        let lexer = Token::lexer(input).spanned();
        CachedLexer {
            iter: lexer,
            cached: VecDeque::new(),
        }
    }

    fn view(&mut self) -> LexerView {
        LexerView {
            ix: 0,
            iter: Box::new(self),
        }
    }

    pub fn parse_program(&mut self) -> Option<Program> {
        parse_program(self.view())
    }
}

type NextView<'a> =
    Box<&'a mut dyn Lookahead<Item = (Token, Range<usize>), LookedItem = (Token, Range<usize>)>>;

struct LexerView<'a> {
    /// offset from iter position
    ix: usize,
    /// parent lookahead iterator
    iter: NextView<'a>,
}

impl Iterator for LexerView<'_> {
    type Item = (Token, Range<usize>);

    fn next(&mut self) -> Option<Self::Item> {
        let res = self.iter.lookahead(self.ix);
        self.ix += 1;
        res
    }
}

impl Lookahead for LexerView<'_> {
    type LookedItem = (Token, Range<usize>);

    fn lookahead(&mut self, times: usize) -> Option<Self::LookedItem> {
        self.iter.lookahead(self.ix + times)
    }

    fn digest(&mut self, times: usize) {
        self.ix += times;
    }
}

impl<'a, 'b> LexerView<'a> {
    fn commit(self) {
        self.iter.digest(self.ix);
    }

    fn peek(&mut self) -> Option<(Token, Range<usize>)> {
        self.lookahead(0)
    }

    fn view(&'b mut self) -> LexerView {
        LexerView {
            ix: 0,
            iter: Box::new(self),
        }
    }
}

fn parse_program(mut view: LexerView<'_>) -> Option<Program> {
    let mut blocks = Vec::new();
    while let Some((token, _)) = view.peek() {
        match token {
            Token::GeneratorLine => {
                blocks.push(Blocks::Generator(parse_generator_block(view.view())?));
            }
            Token::SettingLine => {
                blocks.push(Blocks::Setting(parse_setting_block(view.view())?));
            }
            Token::AtTimeLine => {
                blocks.push(Blocks::TimeDefinition(parse_timeanchor_block(view.view())?));
            }
            Token::FillerLine => {
                consume_filler_lines(view.view());
            }
            _ => return None,
        }
    }
    view.commit();
    Some(Program(blocks))
}

fn parse_timeanchor_block(mut view: LexerView<'_>) -> Option<TimeAnchorBlock> {
    let (Token::AtTimeLine, s) = view.next()? else {
        return None;
    };
    let timeref = parse_time_ref(view.view())?;
    let freq = if let (Token::TimeStepFreq(f), _) = view.peek()? {
        view.next();
        // f is of format "1Hz" or "1kHz"
        let scale = f.chars().skip_while(|c| c.is_numeric()).collect::<String>();
        let freq = f
            .chars()
            .take_while(|c| c.is_numeric())
            .collect::<String>()
            .parse::<f64>()
            .ok()?;
        let freq = match scale.as_str() {
            "Hz" => freq,
            "kHz" => freq * 1e3,
            "MHz" => freq * 1e6,
            "GHz" => freq * 1e9,
            "THz" => freq * 1e12,
            _ => return None,
        };
        Some(freq)
    } else {
        None
    };
    consume_new_line(view.view());
    view.commit();
    Some(TimeAnchorBlock {
        span: s.start..s.end,
        timeref,
        freq,
    })
}

fn parse_generator_block(mut view: LexerView<'_>) -> Option<GeneratorBlock> {
    let (Token::GeneratorLine, s) = view.next()? else {
        return None;
    };
    let generator = parse_generator(view.view())?;
    consume_new_line(view.view());
    let mut parameters = Vec::new();
    while let (Token::ParameterLine, _) = view.peek()? {
        view.next();
        parameters.push(parse_expr_parameter(view.view())?);
        consume_new_line(view.view());
    }
    view.commit();
    Some(GeneratorBlock {
        span: s.start..parameters.last().unwrap().span.end,
        kind: generator.kind,
        parameters,
    })
}

fn parse_setting_block(mut view: LexerView<'_>) -> Option<SettingBlock> {
    let (Token::SettingLine, s) = view.next()? else {
        return None;
    };
    let setting = parse_setting(view.view())?;
    consume_new_line(view.view());
    let mut settings = Vec::new();
    while let (Token::ParameterLine, _) = view.peek()? {
        view.next();
        settings.push(parse_num_parameter(view.view())?);
        consume_new_line(view.view());
    }
    view.commit();
    Some(SettingBlock {
        span: s.start..settings.last().unwrap().span.end,
        kind: setting.name,
        settings,
    })
}

fn parse_time_ref(mut view: LexerView<'_>) -> Option<TimeRef> {
    if let (Token::TimeRef(t), r) = view.next()? {
        view.commit();
        Some(TimeRef { span: r, offset: t })
    } else {
        None
    }
}

fn parse_setting(mut view: LexerView<'_>) -> Option<Setting> {
    if let (Token::BracketRef(set), span) = view.next()? {
        if set == "GPS" {
            view.commit();
            return Some(Setting {
                span,
                name: SettingKind::GPS,
            });
        }
    }
    None
}

fn parse_generator(mut view: LexerView<'_>) -> Option<Generator> {
    if let (Token::BracketRef(gen), span) = view.next()? {
        if gen == "position.parametric" {
            view.commit();
            return Some(Generator {
                span,
                kind: GeneratorKind::PositionParametric,
            });
        }
    };
    None
}

fn parse_num_parameter(mut view: LexerView<'_>) -> Option<NumberParameter> {
    let (Token::Parameter(x), r) = view.next()? else {
        return None;
    };
    let number = parse_number(view.view())?;
    view.commit();
    Some(NumberParameter {
        span: r.start..number.span.end,
        name: x,
        value: number,
    })
}

fn parse_expr_parameter(mut view: LexerView<'_>) -> Option<ExprParameter> {
    let (Token::Parameter(x), r) = view.next()? else {
        return None;
    };
    let expr = parse_expr(view.view())?;
    view.commit();
    Some(ExprParameter {
        span: r.start..expr.span().end,
        name: x,
        value: expr,
    })
}

fn parse_expr(mut view: LexerView<'_>) -> Option<TimeExpr> {
    let res = parse_sum(view.view());
    view.commit();
    res
}

fn parse_sum(mut view: LexerView<'_>) -> Option<TimeExpr> {
    let lhs = parse_product(view.view())?;
    consume_any_space(view.view());
    let res = match view.peek() {
        Some((Token::OpAdd, _)) => {
            view.next();
            consume_any_space(view.view());
            let rhs = parse_sum(view.view())?;
            TimeExpr::Op(Box::new(Operation {
                span: lhs.span().start..rhs.span().end,
                op: Operator::Add,
                lhs: Box::new(lhs),
                rhs: Box::new(rhs),
            }))
        }
        Some((Token::OpSub, _)) => {
            view.next();
            consume_any_space(view.view());
            let rhs = parse_sum(view.view())?;
            TimeExpr::Op(Box::new(Operation {
                span: lhs.span().start..rhs.span().end,
                op: Operator::Sub,
                lhs: Box::new(lhs),
                rhs: Box::new(rhs),
            }))
        }
        _ => lhs,
    };
    view.commit();
    Some(res)
}

fn parse_product(mut view: LexerView<'_>) -> Option<TimeExpr> {
    let lhs = parse_power(view.view())?;
    consume_any_space(view.view());
    let res = match view.peek() {
        Some((Token::OpMul, _)) => {
            view.next();
            consume_any_space(view.view());
            let rhs = parse_product(view.view())?;
            TimeExpr::Op(Box::new(Operation {
                span: lhs.span().start..rhs.span().end,
                op: Operator::Mul,
                lhs: Box::new(lhs),
                rhs: Box::new(rhs),
            }))
        }
        Some((Token::OpDiv, _)) => {
            view.next();
            consume_any_space(view.view());
            let rhs = parse_product(view.view())?;
            TimeExpr::Op(Box::new(Operation {
                span: lhs.span().start..rhs.span().end,
                op: Operator::Div,
                lhs: Box::new(lhs),
                rhs: Box::new(rhs),
            }))
        }
        Some((Token::Time, _)) => {
            let rhs = parse_power(view.view())?;
            TimeExpr::Op(Box::new(Operation {
                span: lhs.span().start..rhs.span().end,
                op: Operator::Mul,
                lhs: Box::new(lhs),
                rhs: Box::new(rhs),
            }))
        }
        _ => lhs,
    };
    view.commit();
    Some(res)
}

fn parse_power(mut view: LexerView<'_>) -> Option<TimeExpr> {
    let lhs = parse_expr_value(view.view())?;
    consume_any_space(view.view());
    let res = if let Some((Token::OpPow, _)) = view.peek() {
        view.next();
        consume_any_space(view.view());
        let rhs = parse_power(view.view())?;
        TimeExpr::Op(Box::new(Operation {
            span: lhs.span().start..rhs.span().end,
            op: Operator::Pow,
            lhs: Box::new(lhs),
            rhs: Box::new(rhs),
        }))
    } else {
        lhs
    };
    view.commit();
    Some(res)
}

fn parse_expr_value(mut view: LexerView<'_>) -> Option<TimeExpr> {
    if let Some(number) = parse_number(view.view()) {
        view.commit();
        return Some(TimeExpr::Num(number));
    }
    if let Some(time_var) = parse_time_var(view.view()) {
        view.commit();
        return Some(TimeExpr::TimeVar(time_var));
    }
    if let Some(group) = parse_group(view.view()) {
        view.commit();
        return Some(TimeExpr::Group(group));
    }

    None
}

fn parse_number(mut view: LexerView<'_>) -> Option<Number> {
    let (token, range) = view.next()?;
    let res = match token {
        Token::Num(n) => Number {
            span: range,
            value: n,
        },
        Token::OpSub => {
            let n = parse_number(view.view())?;
            Number {
                span: range.start..n.span.end,
                value: -n.value,
            }
        }
        _ => return None,
    };
    view.commit();
    Some(res)
}

fn parse_group(mut view: LexerView<'_>) -> Option<GroupExpr> {
    if let (Token::LParen, s) = view.next()? {
        let expr = parse_expr(view.view())?;
        if let (Token::RParen, e) = view.next()? {
            view.commit();
            return Some(GroupExpr {
                span: s.start..e.end,
                value: Box::new(expr),
            });
        }
    }
    None
}

fn parse_time_var(mut view: LexerView<'_>) -> Option<TimeVar> {
    if let (Token::Time, r) = view.next()? {
        view.commit();
        Some(TimeVar { span: r })
    } else {
        None
    }
}

fn consume_filler_lines(mut view: LexerView<'_>) {
    while let Some((Token::FillerLine, _)) = view.peek() {
        view.next();
        consume_new_line(view.view());
    }
    view.commit();
}

fn consume_any_space(mut view: LexerView<'_>) {
    while let Some((Token::Spaces, _)) = view.peek() {
        view.next();
    }
    view.commit();
}

fn consume_new_line(mut view: LexerView<'_>) -> Option<()> {
    match view.next()? {
        (Token::Newline, _) => {
            view.commit();
            Some(())
        }
        _ => None,
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parse_num() {
        let mut lex = CachedLexer::new("10.23");
        assert_eq!(parse_number(lex.view()).unwrap().value, 10.23);

        let mut lex = CachedLexer::new("-123");
        assert_eq!(parse_number(lex.view()).unwrap().value, -123.0);

        let mut lex = CachedLexer::new("0");
        assert_eq!(parse_number(lex.view()).unwrap().value, 0.0);

        let mut lex = CachedLexer::new("-0.0");
        assert_eq!(parse_number(lex.view()).unwrap().value, 0.0);

        let mut lex = CachedLexer::new("42 12.2");
        assert_eq!(parse_number(lex.view()).unwrap().value, 42.0);
        lex.next();
        assert_eq!(parse_number(lex.view()).unwrap().value, 12.2);

        let mut lex = CachedLexer::new("t12.2");
        assert!(parse_number(lex.view()).is_none());
        lex.next();
        assert_eq!(parse_number(lex.view()).unwrap().value, 12.2);
    }

    #[test]
    fn test_parse_expr() {
        let mut lex = CachedLexer::new("(3.4+2)*3.1");
        parse_expr(lex.view()).unwrap();

        let mut lex = CachedLexer::new("3.4+2*3.1");
        parse_expr(lex.view()).unwrap();

        let mut lex = CachedLexer::new("(3^2+1) / 2");
        parse_expr(lex.view()).unwrap();

        let mut lex = CachedLexer::new("2t^2+1.5t-3");
        parse_expr(lex.view()).unwrap();
    }
}
